/* -*- c++ -*-

This is Picprog, Microchip PIC programmer software for the serial port device.
Copyright  1997,2002,2003,2004 Jaakko Hyvtti

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

The author may be contacted at:

Email: Jaakko.Hyvatti@iki.fi
URL:   http://www.iki.fi/hyvatti/
Phone: +358 40 5011222

Please send any suggestions, bug reports, success stories etc. to the
Email address above.

*/

#include <iostream>
#include <iomanip>

#include <sysexits.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include <unistd.h>

#include "hexfile.h"

using namespace std;

/*
  The exact programmable memory type, prom/eprom/eeprom/flash, does
  not affect the operations of the program.  Therefore I have not
  checked if every chip has the correct type specified.

  If rom type is specified, this program only reads the memory and
  does not atempt to burn it.

  Structure fields:

  - Name, used in command line option to select the device.
  - Program memory size in words.
  - Number of preserved words at the end of program memory, for example
    oscillator calibration word.  I have not checked all datasheets
    for which devices have one, only some.  Please check the datasheet
    before you overwrite these configuration words!  As a precaution,
    please read them first off the chip and write down somewhere!
  - configuration word preserved bits mask
  - Word bit length.  Only 14 bit ones supported.
  - Program memory type.
  - Non-volatile data memory size.
  - Non-volatile data memory type.
  - location 0x2006 device id (-1 == device has no known id)
*/

const struct hexfile::devinf hexfile::deviceinfo [] = {

  // 16x8x family

  {"pic16c84", 1024, 0, 0, 14, eeprom, 64, eeprom, -1}, // no OSCCAL
  {"pic16cr83", 512, 0, 0, 14, rom, 64, eeprom, -1},
  {"pic16cr84", 1024, 0, 0, 14, rom, 64, eeprom, -1},
  {"pic16f83", 512, 0, 0, 14, flash, 64, eeprom, -1},
  {"pic16f84", 1024, 0, 0, 14, flash, 64, eeprom, -1}, // no OSCCAL
  {"pic16f84a", 1024, 0, 0, 14, flash, 64, eeprom, 0x0560},
  {"pic16f87", 4096, 0, 0, 14, flash5, 256, eeprom, 0x0720},
  {"pic16f88", 4096, 0, 0, 14, flash5, 256, eeprom, 0x0760},

  // 16c6x family

  {"pic16c61",  1024, 0, 0, 14, prom, 0, rom, -1}, // ?
  {"pic16c62",  2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c62a", 2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c62b", 2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c63",  4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c63a", 4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c64",  2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c64a", 2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c65",  4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c65a", 4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c65b", 4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c66",  8192, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c66a", 8192, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c67",  8192, 0, 0, 14, prom, 0, rom, -1},
  {"pic16cr62", 2048, 0, 0, 14, rom, 0, rom, -1},
  {"pic16cr63", 4096, 0, 0, 14, rom, 0, rom, -1},
  {"pic16cr64", 2048, 0, 0, 14, rom, 0, rom, -1},
  {"pic16cr65", 4096, 0, 0, 14, rom, 0, rom, -1},

  // 16c62x family

  {"pic16c620", 512, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c620a", 512, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16cr620a", 512, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c621", 1024, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c621a", 1024, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c622", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c622a", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16f627",  1024, 0, 0, 14, flash, 128, eeprom, 0x07a0}, // no OSCCAL
  {"pic16f627a", 1024, 0, 0, 14, flash4, 128, eeprom, 0x1040}, // no OSCCAL
  {"pic16f628",  2048, 0, 0, 14, flash, 128, eeprom, 0x07c0}, // no OSCCAL
  {"pic16f628a", 2048, 0, 0, 14, flash4, 128, eeprom, 0x1060}, // no OSCCAL
  {"pic16f648a", 4096, 0, 0, 14, flash4, 128, eeprom, 0x1100}, // no OSCCAL

  // 16ce62x family

  {"pic16ce623", 512, 0, 0, 14, eprom, 128, eeprom, -1},
  {"pic16ce624", 1024, 0, 0, 14, eprom, 128, eeprom, -1},
  {"pic16ce625", 2048, 0, 0, 14, eprom, 128, eeprom, -1},

  // 16c64x, 16c66x families

  {"pic16c641", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c642", 4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c661", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c662", 4096, 0, 0, 14, eprom, 0, rom, -1},

  // 16c7x, 16c77x families

  {"pic16c71",  1024, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c710", 512, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c711", 1024, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c712", 1024, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c715", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c716", 2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c717", 2048, 0, 0, 14, eprom, 0, rom, -1},

  {"pic16c72",  2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c72a",  2048, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16cr72",  2048, 0, 0, 14, rom, 0, rom, -1},
  {"pic16c73",  4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c73a", 4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c73b", 4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c74",  4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c74a", 4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c74b", 4096, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c76", 8192, 0, 0, 14, eprom, 0, rom, -1},
  {"pic16c77", 8192, 0, 0, 14, eprom, 0, rom, -1},

  // 16f7x family

  {"pic16f72", 4096, 0, 0, 14, flash2, 0, rom, 0x00a0},
  {"pic16f73", 4096, 0, 0, 14, flash2, 0, rom, 0x0600},
  {"pic16f74", 4096, 0, 0, 14, flash2, 0, rom, 0x0620},
  {"pic16f76", 8192, 0, 0, 14, flash2, 0, rom, 0x0640},
  {"pic16f77", 8192, 0, 0, 14, flash2, 0, rom, 0x0660},

  // 16c43x family
  
  {"pic16c432", 2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c433", 2048, 1, 0, 14, prom, 0, rom, -1},

  // 16c78x family

  {"pic16c781", 1024, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c782", 2048, 0, 0, 14, prom, 0, rom, -1},

  // 16c7x5 family

  {"pic16c745", 8192, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c765", 8192, 0, 0, 14, prom, 0, rom, -1},

  // 16c77x family

  {"pic16c770", 2048, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c771", 4096, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c773", 8192, 0, 0, 14, prom, 0, rom, -1},
  {"pic16c774", 8192, 0, 0, 14, prom, 0, rom, -1},
  
  // 16f87x family

  {"pic16f870", 2048, 0, 0, 14, flash, 64, eeprom, 0x0d00},
  {"pic16f871", 2048, 0, 0, 14, flash, 64, eeprom, 0x0d20},
  {"pic16f872", 2048, 0, 0, 14, flash, 64, eeprom, 0x08e0},
  {"pic16f873", 4096, 0, 0, 14, flash, 128, eeprom, 0x0960},
  {"pic16f873a", 4096, 0, 0, 14, flash3, 128, eeprom, 0x0e40},
  {"pic16f874", 4096, 0, 0, 14, flash, 128, eeprom, 0x0920},
  {"pic16f874a", 4096, 0, 0, 14, flash3, 128, eeprom, 0x0e60},
  {"pic16f876", 8192, 0, 0, 14, flash, 256, eeprom, 0x09e0},
  {"pic16f876a", 8192, 0, 0, 14, flash3, 256, eeprom, 0x0e00},
  {"pic16f877", 8192, 0, 0, 14, flash, 256, eeprom, 0x09a0},
  {"pic16f877a", 8192, 0, 0, 14, flash3, 256, eeprom, 0x0e20},

  {"pic16f818", 1024, 0, 0, 14, flash5, 128, eeprom, 0x04c0},
  {"pic16f819", 2048, 0, 0, 14, flash5, 128, eeprom, 0x04e0},

  {"pic16c923", 4096, 0, 0, 14, eprom, 0, eprom, -1},
  {"pic16c924", 4096, 0, 0, 14, eprom, 0, eprom, -1},

  {"pic16f630", 1024, 1, 0x3000, 14, flash4, 128, eeprom, 0x10c0},
  {"pic16f676", 1024, 1, 0x3000, 14, flash4, 128, eeprom, 0x10e0},

  // 12

  {"pic12c508",   512, 1, 0, 12, eprom,   0, rom, -1},
  {"pic12c508a",  512, 1, 0, 12, eprom,   0, rom, -1},
  {"pic12ce518",  512, 0, 0, 12, eprom,  16, eeprom, -1},
  {"pic12c509",  1024, 1, 0, 12, eprom,   0, rom, -1},
  {"pic12c509a", 1024, 0, 0, 12, eprom,   0, rom, -1},
  {"pic12ce519", 1024, 0, 0, 12, eprom,  16, eeprom, -1},
  {"pic12cr509a",1024, 0, 0, 12, rom,     0, rom, -1},
  {"pic12c671",  1024, 1, 0, 14, eprom,   0, rom, 0x0500},
  {"pic12c672",  2048, 1, 0, 14, eprom,   0, rom, -1},
  {"pic12ce673", 1024, 1, 0, 14, eprom,  16, eeprom, -1},
  {"pic12ce674", 2048, 1, 0, 14, eprom,  16, eeprom, -1},
  {"pic12f629",  1024, 1, 0x3000, 14, flash4, 128, eeprom, 0x0f80},
  {"pic12f675",  1024, 1, 0x3000, 14, flash4, 128, eeprom, 0x0fc0},

  {"pic16c505",  1024, 1, 0, 12, eprom,   0, rom, -1},

#if 0
  // The 18f series is not supported, they are just listed here
  // to hold the id numbers.
  {"pic18f242", 16 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0480},
  {"pic18f248", 16 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0800},
  {"pic18f252", 32 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0400},
  {"pic18f258", 32 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0840},
  {"pic18f442", 16 * 1024, 0, 0, 14, flash, 256, eeprom, 0x04A0},
  {"pic18f448", 16 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0820},
  {"pic18f452", 32 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0420},
  {"pic18f458", 32 * 1024, 0, 0, 14, flash, 256, eeprom, 0x0860},
#endif
};

static int got_signal = 0;

void
term_handler (int a)
{
  got_signal = 1;
  signal (a, term_handler);
}

int
hexfile::load (const char *name)
{
  ifstream f(name);
  char buf [128];
  int line = 0;
  enum formats format = unknown;
  int addr, words, check, sum, word, i;
  int e;

  if (!f) {
    e = errno;
    cerr << name << ":unable to load hexfile:" << strerror (e) << endl;
    return EX_NOINPUT;
  }
  while (f.get (buf, sizeof (buf))) {
    line++;
    char c;
    if (f.get (c) && '\n' != c) {
      cerr << name << ':' << line << ":long input line" << endl;
      return EX_DATAERR;
    }
    int len = strlen (buf);
    while (len && isspace (buf [len-1]))
      len--;
    buf [len] = '\0';
    if (!strcmp (buf, ":00000001FF")) // eof
      return EX_OK;

    if (!len)
      continue; // empty line, I do not want to whine about it

    i = len - 1;
    while (i && isxdigit (buf [i]))
      i--;

    if (i || ':' != buf [0] || 3 != (len & 3) || len < 15
	|| '0' != buf [7] || '0' != buf [8]) {
      cerr << name << ':' << line << ":invalid input line." << endl
	   << "Are you sure this is an 8 or 16 bit intel hex file?" << endl;
      return EX_DATAERR;
    }
    check = strtol (buf + len - 2, 0, 16);
    buf [7] = '\0';
    addr = strtol (buf + 3, 0, 16);
    buf [3] = '\0';
    words = strtol (buf + 1, 0, 16);
    if (unknown == format) {
      if (words * 4 + 11 == len)
	format = ihx16;
      else if (words * 2 + 11 == len)
	format = ihx8m;
      else {
	cerr << name << ':' << line <<
	  ":unknown input format, only ihx8m and ihx16 accepted" << endl;
	return EX_DATAERR;
      }
    }
    if (words * (ihx16 == format ? 4 : 2) + 11 != len) {
      cerr << name << ':' << line << ":line length mismatch:"
	   << (ihx16 == format ? "ihx16 " : "ihx8m ")
	   << words * (ihx16 == format ? 4 : 2) + 11 << " != " << len << endl;
      return EX_DATAERR;
    }
    sum = words + addr + (addr >> 8);
    if (ihx8m == format) {
      if ((words & 1) || (addr & 1)) {
	cerr << name << ':' << line << ":odd address or number of words."
	     << endl;
	return EX_DATAERR;
      }
      words /= 2;
      addr /= 2;
    }

    if (!((addr >= 0 && addr+words <= deviceinfo [dev].prog_size)
	  || (addr >= 0x2000 && addr+words <= 0x2004)
	  || (addr == 0x2007 && words == 1)
	  || (addr >= 0x2100 && addr+words <= 0x2100 + deviceinfo [dev].data_size))) {
      cerr << name << ':' << line << ":invalid address " << hex
	   << setw(4) << setfill('0') << addr << dec
	   << ", possibly not hex file for correct pic type?" << endl;
      return EX_DATAERR;
    }

    while (words--) {
      buf [13 + words * 4] = '\0';
      word = strtol (buf + 9 + words * 4, 0, 16);
      if (ihx8m == format)
	word = (word & 0xff) << 8 | word >> 8;
      sum += word + (word >> 8);
      pgm [addr + words] = word;
    }
    if (sum + check & 0xff) {
      cerr << name << ':' << line << ":checksum mismatch" << endl;
      return EX_DATAERR;
    }
  }
  e = errno;
  if (!f.eof ()) {
    cerr << name << ':' << line << ':' << strerror (e) << ":" << endl;
    return EX_IOERR;
  }
  cerr << name << ':' << line << ":warning:unexpected eof" << endl;
  return EX_OK;
}

void
hexfile::save_line (ofstream& f, int begin, int len, enum hexfile::formats format) const
{
  int p_begin, p_len, sum, i;

  if (ihx8m == format) {
    p_begin = begin * 2;
    p_len = len * 2;
  } else {
    p_begin = begin;
    p_len = len;
  }

  f << ':' << setw (2) << p_len << setw (4) << p_begin << "00";

  sum = p_len + p_begin + (p_begin >> 8);
  for (i = begin; i < begin + len; i++) {
    int word = pgm [i];
    if (ihx8m == format)
      word = (word & 0xff) << 8 | (word & 0xff00) >> 8;
    f << setw (4) << word;
    sum += word + (word >> 8);
  }

  f << setw (2) << (-sum & 0xff) << endl;
}

void
hexfile::save_region (ofstream& f, int addr, int end, enum hexfile::formats format, bool skip_ones) const
{
  int len;

  for (;;) {
    while (addr < end
	   && (-1 == pgm [addr]
	       || (skip_ones
		   && (addr >= 0x2100 && addr < 0x2100 + deviceinfo [dev].data_size ? 0xff : 0x3fff)
		      == pgm [addr])))
      addr++;
    if (addr >= end)
      break;
    len = 1;
    while (len < 8 && addr+len < end
	   && !(-1 == pgm [addr+len]
		|| (skip_ones
		    && (addr+len>= 0x2100 && addr+len < 0x2100 + deviceinfo [dev].data_size ? 0xff : 0x3fff)
		       == pgm [addr+len])))
      len++;
    save_line (f, addr, len, format);
    addr += len;
  }
}

int
hexfile::save (const char *name, enum hexfile::formats format, bool skip_ones) const
{
  ofstream f (name);
  int e;

  if (!f) {
    e = errno;
    cerr << name << ":unable to open save file:" << strerror (e) << endl;
    return EX_IOERR;
  }
  f << hex << setfill ('0') << setiosflags (ios::uppercase);

  save_region (f, 0, deviceinfo [dev].prog_size, format, skip_ones);
  save_region (f, 0x2000, 0x2004, format, skip_ones);
  save_region (f, 0x2007, 0x2008, format, skip_ones);
  if (deviceinfo [dev].data_size)
    save_region (f, 0x2100, 0x2100 + deviceinfo [dev].data_size, format, skip_ones);
  f << ":00000001FF" << endl;
  return EX_OK;
}

/*
It will read both IHX16 and IHX8M (auto-detection), program ID and
fuse locations and data memory.  It verifies location before and after
programming, so unnecessary programming is avoided and errors are
detected.  Code protection is detected before programming, and bulk
erase is executed if other than ID locations are programmed if CP is
enabled.  */

int
hexfile::program_location (picport& pic, int addr, bool data) const
{
  int retval;

  if (-1 == pgm [addr]
      || (retval = pic.command (data ? picport::data_from_data
				: picport::data_from_prog)) == pgm [addr])
    return NOT_PROGRAMMED;

  if (-1 == retval) {
    cerr << pic.port() << ':' << hex << setfill ('0') << setw (4) << addr
	 << dec << ":unable to read pic while programming" << endl;
    return EX_IOERR;
  }
  pic.command (data ? picport::data_for_data : picport::data_for_prog,
	       pgm [addr]);
  switch (deviceinfo [dev].prog_type) {
  case flash2: // pic16f77
    pic.command (picport::beg_prog);
    usleep (10000);
    pic.command (picport::end_prog);
    break;
  case flash3: // pic16f876a
  case flash5: // pic16f88
    pic.command (picport::beg_prog_only);
    usleep (10000);
    pic.command (picport::end_prog_only);
    break;
  default: // flash, flash4, eeprom
    pic.command (picport::beg_prog);
    usleep (10000);
  }

  // verify, but do not verify fuses if Code Protect bit is cleared!

  int read_val;
  if (pgm [addr] != (read_val =
		     pic.command (data ? picport::data_from_data
				  : picport::data_from_prog))) {
    cerr << pic.port() << ':' << hex << setw (4) << setfill ('0') << addr
	 << ": programmed=" << setw (4) << setfill ('0') << pgm [addr]
	 << ", read=" << setw (4) << setfill ('0') << read_val
	 << dec << ":unable to verify pic while programming." << endl;
    if (pic.address () != 0x2007) {
      cerr << "Is code protection enabled, or does the chip need to be " << endl
	   << "erased completely before programming?" << endl
	   << "Use --erase option to disable code protection." << endl;
      return EX_IOERR;
    } else {
      cerr << "This is the configuration word, which often has hardwired" << endl
	   << "bits and therefore does not verify.  It also has the code" << endl
	   << "protection bits, and if they were programmed to enabled" << endl
	   << "state, verification fails.  Therefore this error is ignored." << endl;
    }
  }
  return EX_OK;
}

typedef void (*sig_type)(int);

void hexfile::reset_code_protection (picport& pic)
{
  switch (deviceinfo [dev].prog_type) {
  case flash2: // pic16f77
    pic.command (picport::load_conf, 0x3fff);
    pic.command (picport::erase_prog);
    break;
  case flash3: // pic16f876a
  case flash5: // pic16f88
    pic.command (picport::load_conf, 0x3fff);
    pic.command (picport::chip_erase);
    break;
  case flash4: // pic16f628a
    pic.command (picport::load_conf, 0x3fff);
    pic.command (picport::erase_prog);
    usleep (50000);
    pic.command (picport::erase_data);
    break;
  default: // eeprom, flash
    pic.command (picport::load_conf, 0x3fff);
    for (int i = 0; i < 7; i++)
      pic.command (picport::inc_addr);
    assert (0x2007 == pic.address ());
    pic.command (picport::command1);
    pic.command (picport::command7);
    pic.command (picport::beg_prog);
    usleep (20000);
    pic.command (picport::command1);
    pic.command (picport::command7);

    // On my pic16f628, the above leaves data memory not erased,
    // if the code protection was not on.
    pic.command (picport::data_for_data, 0x3fff);
    pic.command (picport::erase_data);
    pic.command (picport::beg_prog);
  }

  usleep (50000);
  pic.reset ();
}

int
hexfile::program (const char *port, bool reset, bool nopreserve)
{
  int retval;

  sig_type save_t, save_q, save_i;
  save_t = signal (SIGTERM, term_handler);
  save_q = signal (SIGQUIT, term_handler);
  save_i = signal (SIGINT, term_handler);

  cout << "Device " << deviceinfo [dev].name
       << ", program memory: " << deviceinfo [dev].prog_size
       << ", data memory: " << deviceinfo [dev].data_size << "."
       << endl;
  if ((rom == deviceinfo [dev].data_type || 0 == deviceinfo [dev].data_size)
      && (rom == deviceinfo [dev].prog_type || 0 == deviceinfo [dev].prog_size)) {
    cerr << "This type of device has no programmable memory." << endl;
    return EX_USAGE;
  }

  {
    picport pic (port);

    if (deviceinfo [dev].prog_preserved && !nopreserve) {
      if (!reset) {
	// Chip is not erased, so we just need to avoid programming
	// the reserved locations.
	for (int addr = deviceinfo [dev].prog_size - deviceinfo [dev].prog_preserved;
	     addr < deviceinfo [dev].prog_size;
	     ++addr) {
	  cout << "Calibration word at "
	       << hex << setfill('0') << setw(4) << addr << dec
	       << " not programmed";
	  if (-1 != pgm [addr]) {
	    pgm [addr] = -1;
	    cout << " (value in input file ignored)";
	  }
	  cout << endl;
	}
      } else {
	// We are erasing the chip. Read OSCCAL and other reserved
	// words off the device first.
	for (;;) {
	  if (pic.address () >= deviceinfo [dev].prog_size - deviceinfo [dev].prog_preserved) {
	    int value = pic.command (picport::data_from_prog);
	    if (-1 == value) {
	      cerr << port << ':'
		   << hex << setfill ('0') << setw (4) << pic.address () << dec
		   << ":unable to read pic calibration words" << endl;
	      return EX_IOERR;
	    }
	    cout << "Calibration word at "
		 << hex << setfill('0') << setw(4) << pic.address()
		 << " preserved as "
		 << setfill('0') << setw(4) << value << dec;
	    if (-1 != pgm [pic.address ()])
	      cout << " (value in input file ignored)";
	    cout << endl;
	    pgm [pic.address ()] = value;
	  }
	  if (pic.address () + 1 >= deviceinfo [dev].prog_size)
	    break;
	  pic.command (picport::inc_addr);
	} // for
      }
    }
    if (deviceinfo [dev].config_mask && !nopreserve) {
      if (reset || -1 != pgm [0x2007]) {
	pic.command (picport::load_conf, 0);
	for (int i = 0; i < 7; ++i)
	  pic.command (picport::inc_addr);
	assert (0x2007 == pic.address());
	int value;
	if (-1 == (value = pic.command (picport::data_from_prog))) {
	  cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	       << ":unable to read pic calibration bits" << endl;
	  return EX_IOERR;
	}
	value &= deviceinfo [dev].config_mask;
	cout << "Calibration bits "
	     << hex << setfill('0') << setw(4) << deviceinfo [dev].config_mask
	     << " in configuration word preserved as "
	     << setfill('0') << setw(4) << value << dec << endl;
	if (-1 == pgm [pic.address ()])
	  pgm [pic.address ()] = 0x3fff;
	pgm [pic.address ()] = pgm [pic.address ()] & ~deviceinfo [dev].config_mask | value;
      } else {
	cout << "Calibration bits "
	     << hex << setfill('0') << setw(4) << deviceinfo [dev].config_mask
	     << dec << " in configuration word not programmed" << endl;
      }
    }

    if (reset) {
      reset_code_protection (pic);
      cout << "Erased and removed code protection." << endl;
    }
    if (pic.address ())
      pic.reset ();

    int count;
    if (rom == deviceinfo [dev].prog_type || 0 == deviceinfo [dev].prog_size) {
      cout << "Skipped burning program memory," << endl;
    } else {
      cout << "Burning program memory," << flush;
      count = 0;
      while (pic.address () < deviceinfo [dev].prog_size) {
	retval = program_location (pic, pic.address (), false);
	if (EX_OK == retval)
	  ++count;
	else if (NOT_PROGRAMMED != retval)
	  return retval;
	pic.command (picport::inc_addr);
	if (got_signal) {
	  cerr << "Exiting." << endl;
	  return EX_UNAVAILABLE;
	}
      }
      cout << " " << count << " locations," << endl;
    }

    if (rom == deviceinfo [dev].data_type || 0 == deviceinfo [dev].data_size) {
      cout << "skipped burning data memory," << endl;
    } else {
      cout << "burning data memory," << flush;
      count = 0;
      while (pic.address () < deviceinfo [dev].prog_size + deviceinfo [dev].data_size) {
	retval = program_location (pic,
				   pic.address () + 0x2100 - deviceinfo [dev].prog_size,
				   true);
	if (EX_OK == retval)
	  ++count;
	else if (NOT_PROGRAMMED != retval)
	  return retval;
	pic.command (picport::inc_addr);
	if (got_signal) {
	  cerr << "Exiting." << endl;
	  return EX_UNAVAILABLE;
	}
      }
      cout << " " << count << " locations," << endl;
    }

    cout << "burning id words," << flush;
    count = 0;
    pic.command (picport::load_conf, 0x3fff); // dummy value
    while (pic.address () < 0x2004) {
      retval = program_location (pic, pic.address (), false);
      if (EX_OK == retval)
	++count;
      else if (NOT_PROGRAMMED != retval)
	return retval;
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting." << endl;
	return EX_UNAVAILABLE;
      }
    }
    cout << " " << count << " locations," << endl;

    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);

    cout << "burning fuses," << flush;
    count = 0;
    retval = program_location (pic, pic.address (), false);
    if (EX_OK == retval)
      ++count;
    else if (NOT_PROGRAMMED != retval)
      return retval;
    cout << " " << count << " locations," << endl;
    cout << "done." << endl;
  }  // scope of pic

  signal (SIGTERM, save_t);
  signal (SIGQUIT, save_q);
  signal (SIGINT, save_i);
  if (got_signal) {
    cerr << "Exiting." << endl;
    return EX_UNAVAILABLE;
  }

  return EX_OK;
}


int
hexfile::read (const char *port)
{

  sig_type save_t, save_q, save_i;
  save_t = signal (SIGTERM, term_handler);
  save_q = signal (SIGQUIT, term_handler);
  save_i = signal (SIGINT, term_handler);

  cout << "Device " << deviceinfo [dev].name
       << ", program memory: " << deviceinfo [dev].prog_size;
  if (rom == deviceinfo [dev].prog_type && deviceinfo [dev].prog_size)
    cout << " (rom)";
  cout << ", data memory: " << deviceinfo [dev].data_size;
  if (rom == deviceinfo [dev].data_type && deviceinfo [dev].data_size)
    cout << " (rom)";
  cout << "." << endl;
  {
    picport pic (port);

    if (0 == deviceinfo [dev].prog_size) {
      cout << "Skipped reading program memory," << endl;
    } else {
      cout << "Reading program memory," << endl;
      while (pic.address () < deviceinfo [dev].prog_size) {
	if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
	  cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	       << ":unable to read pic" << endl;
	  return EX_IOERR;
	}
	pic.command (picport::inc_addr);
	if (got_signal) {
	  cerr << "Exiting." << endl;
	  return EX_UNAVAILABLE;
	}
      }
    }

    if (0 == deviceinfo [dev].data_size) {
      cout << "skipped reading data memory," << endl;
    } else {
      cout << "reading data memory," << endl;
      while (pic.address () < deviceinfo [dev].prog_size + deviceinfo [dev].data_size) {
	if (-1 == (pgm [pic.address () + 0x2100-deviceinfo [dev].prog_size]
		   = pic.command (picport::data_from_data))) {
	  cerr << port << ':' << hex << setfill ('0') << setw (4)
	       << pic.address () + 0x2100-deviceinfo [dev].prog_size << dec
	       << ":unable to read pic data memory" << endl;
	  return EX_IOERR;
	}
	pic.command (picport::inc_addr);
	if (got_signal) {
	  cerr << "Exiting." << endl;
	  return EX_UNAVAILABLE;
	}
      }
    }
    pic.command (picport::load_conf, 0);

    cout << "reading id words," << endl;
    while (pic.address () < 0x2004) {
      if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
	cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	     << ":unable to read pic id words" << endl;
	return EX_IOERR;
      }
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting." << endl;
	return EX_UNAVAILABLE;
      }
    }

    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);

    // device id (optional)
    cout << "reading device id,";
    int dev_id;
    if (-1 == (dev_id = pic.command (picport::data_from_prog))
	|| 0x3fff == dev_id) {
      cout << " no id on this device," << endl;
    } else
      cout << " id=" << hex << setfill('0') << setw(4) << (dev_id & 0x3fe0)
	   << ", revision=" << setw(2) << (dev_id & 0x1f) << ","
	   << dec << endl;
    pic.command (picport::inc_addr);

    // fuses
    cout << "reading fuses," << endl;
    if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
      cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	   << ":unable to read pic fuses" << endl;
      return EX_IOERR;
    }
    cout << "done." << endl;
  } // pic scope

  signal (SIGTERM, save_t);
  signal (SIGQUIT, save_q);
  signal (SIGINT, save_i);
  if (got_signal) {
    cerr << "Exiting." << endl;
    return EX_UNAVAILABLE;
  }

  return EX_OK;
}

int
hexfile::setdevice (const char *port, int& d)
{
  if (-1 == d) {
    // Read version information off the chip at address 0x2006
    picport pic (port);
    pic.command (picport::load_conf, 0);
    for (int i = 0; i < 6; ++i)
      pic.command (picport::inc_addr);
    assert (0x2006 == pic.address());
    int value;
    if (-1 == (value = pic.command (picport::data_from_prog))) {
      cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	   << ":unable to read pic device id" << endl;
      return EX_IOERR;
    }
    d = 0;
    for (;;) {
      if (d >= int(sizeof (hexfile::deviceinfo)
		   / sizeof (hexfile::deviceinfo [0]))) {
	d = 0;
	cout << port << ": device id "
	     << hex << setfill ('0') << setw (4) << value << dec
	     << " unknown, defaulting to " << deviceinfo [d].name << endl;
	break;
      }
      if (-1 != deviceinfo [d].device_id
	  && (deviceinfo [d].device_id & 0x3fe0) == (value & 0x3fe0)) {
	cout << port << ": detected " << deviceinfo [d].name
	     << " version " << hex
	     << setfill ('0') << setw (2) << (value & 0x1f) << dec << endl;
	break;
      }
      ++d;
    }
  }
  if (d >= 0
      && d < int(sizeof (hexfile::deviceinfo)
		 / sizeof (hexfile::deviceinfo [0])))
    dev = d;
  else
    d = dev;
  return EX_OK;
}

// Couple of statics in hexfile class.

int
hexfile::find_device (const char *name)
{
  for (unsigned d = 0;
       d < sizeof (hexfile::deviceinfo) / sizeof (struct hexfile::devinf);
       ++d)
    if (!strcasecmp (hexfile::deviceinfo [d].name, name))
      return d;
  return -1;
}

void
hexfile::print_devices ()
{
  for (unsigned d = 0;
       d < sizeof (hexfile::deviceinfo) / sizeof (struct hexfile::devinf);
       ++d) {
    if (deviceinfo [d].prog_bits != 14)
      continue;
    if (d)
      cerr << ", ";
    cerr << deviceinfo [d].name;
    if (-1 != deviceinfo [d].device_id)
      cerr << "*";
  }
  cerr << endl
       << "* = autodetected" << endl;
}
